/*  GNU Moe - My Own Editor
    Copyright (C) 2005, 2006, 2007, 2008, 2009 Antonio Diaz Diaz.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

namespace Encoding {

void base64_encode( const std::string & in, std::string & out ) throw();

bool base64_decode( const std::string & in, std::string & out ) throw();

bool quoted_printable_decode( const std::string & in, std::string & out ) throw();

unsigned char rot13( const unsigned char ch ) throw();

unsigned char rot47( const unsigned char ch ) throw();

bool utf8_decode( const std::string & in, std::string & out ) throw();

} // end namespace Encoding
